/* PulseAudioSourceDataLine.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.AudioFormat.Encoding;
import javax.sound.sampled.Control.Type;

public class PulseAudioSourceDataLine implements SourceDataLine {

	private static final int DEFAULT_BUFFER_SIZE = 1000;
	private static final String PULSEAUDIO_FORMAT_KEY = "PulseAudioFormatKey";

	private String streamName = "Java Stream";

	private EventLoop eventLoop = null;

	private boolean isOpen = false;
	private boolean isPaused = false;

	private AudioFormat[] supportedFormats = null;
	private AudioFormat currentFormat = null;
	private AudioFormat defaultFormat = null;

	private List<LineListener> lineListeners;

	private Control[] controls = null;
	private PulseAudioStreamMuteControl muteControl;
	private PulseAudioStreamVolumeControl volumeControl;
	private boolean muted;
	private float volume;

	private Semaphore semaphore = new Semaphore(0);
	private long currentFramePosition = 0;

	private Stream stream;

	public PulseAudioSourceDataLine(EventLoop eventLoop, AudioFormat[] formats,
			AudioFormat defaultFormat) {

		this.supportedFormats = formats;
		this.eventLoop = eventLoop;
		this.lineListeners = new ArrayList<LineListener>();
		this.volume = PulseAudioVolumeControl.MAX_VOLUME;
		this.defaultFormat = defaultFormat;
		this.currentFormat = defaultFormat;

	}

	protected boolean isMuted() {
		return muted;
	}

	protected void setMuted(boolean value) {
		muted = value;
	}

	protected float getVolume() {
		return this.volume;
	}

	protected void setVolume(float value) {
		this.volume = value;

	}

	public void open(AudioFormat format, int bufferSize)
			throws LineUnavailableException {
		if (isOpen) {
			throw new IllegalStateException("Line is already open");
		}

		// ignore suggested buffer size

		for (AudioFormat myFormat : supportedFormats) {
			if (format.matches(myFormat)) {
				stream = new Stream(eventLoop.getContextPointer(), streamName,
						Stream.Format.valueOf((String) myFormat
								.getProperty(PULSEAUDIO_FORMAT_KEY)),
						(int) format.getSampleRate(), format.getChannels());
				currentFormat = format;
				isOpen = true;
			}
		}
		// no matches found
		if (!isOpen) {
			throw new IllegalArgumentException("Invalid format");
		}

		Stream.StateListener openCloseListener = new Stream.StateListener() {

			@Override
			public void update() {
				if (stream.getState() == Stream.State.READY) {
					fireLineEvent(new LineEvent(PulseAudioSourceDataLine.this,
							LineEvent.Type.OPEN, AudioSystem.NOT_SPECIFIED));
					semaphore.release();
				} else if (stream.getState() == Stream.State.TERMINATED
						|| stream.getState() == Stream.State.FAILED) {
					fireLineEvent((new LineEvent(PulseAudioSourceDataLine.this,
							LineEvent.Type.CLOSE, AudioSystem.NOT_SPECIFIED)));
					semaphore.release();
				}
			}

		};

		stream.addStateListener(openCloseListener);

		synchronized (eventLoop.threadLock) {

			stream.connectForPlayback(null);
		}

		try {
			semaphore.acquire();
		} catch (InterruptedException e) {
			// throw new LineUnavailableException("unable to prepare
			// stream");
		}

		controls = new Control[2];
		volumeControl = new PulseAudioStreamVolumeControl(this);
		controls[0] = volumeControl;
		muteControl = new PulseAudioStreamMuteControl(this);
		controls[1] = muteControl;

	}

	public void open(AudioFormat format) throws LineUnavailableException {
		open(format, DEFAULT_BUFFER_SIZE);

	}

	public void open() throws LineUnavailableException {
		// pick a random format
		if (defaultFormat == null) {
			defaultFormat = new AudioFormat(Encoding.PCM_UNSIGNED, 22050, 8, 2,
					2, AudioSystem.NOT_SPECIFIED, false);
		}

		open(defaultFormat, DEFAULT_BUFFER_SIZE);
	}

	@Override
	public int write(byte[] data, int offset, int length) {

		int frameSize = currentFormat.getFrameSize();
		if (length % frameSize != 0) {
			throw new IllegalArgumentException(
					"amount of data to write does not represent an integral number of frames");
		}
		if (length < 0) {
			throw new IllegalArgumentException("length is negative");
		}

		if (length + offset > data.length) {
			throw new ArrayIndexOutOfBoundsException(length + offset);
		}

		int position = offset;
		int remainingLength = length;
		int availableSize;

		int sizeWritten = 0;

		while (remainingLength != 0) {

			synchronized (eventLoop.threadLock) {
				availableSize = stream.getWritableSize();
				if (availableSize < 0) {
					return sizeWritten;
				}
				if (availableSize > remainingLength) {
					availableSize = remainingLength;
				}
				/* write a little bit of the buffer */
				stream.write(data, position, availableSize);

				sizeWritten += availableSize;
				position += availableSize;
				remainingLength -= availableSize;

			}
		}

		// all the data should have been played by now
		assert (sizeWritten == length);
		currentFramePosition += (sizeWritten / getFormat().getFrameSize());
		/*
		 * FIXME when the stream is flushed() etc, instead of returning length
		 * this should unblock and return the the size of data written so far
		 */
		return sizeWritten;
	}

	public void start() {
		if (isPaused) {
			stream.cork(false);
			isPaused = false;
		}

		/*
		 * for(LineListener l :listeners) { l.update(new LineEvent(this,
		 * LineEvent.Type.START, 0)); }
		 */

	}

	public void stop() {
		synchronized (eventLoop.threadLock) {
			stream.cork(true);
		}
		isPaused = true;

	}

	public void addLineListener(LineListener listener) {
		this.lineListeners.add(listener);
	}

	public void removeLineListener(LineListener listener) {
		this.lineListeners.remove(listener);
	}

	public boolean isOpen() {
		return isOpen;
	}

	public int available() {
		synchronized (eventLoop.threadLock) {
			return stream.getWritableSize();
		}
	};

	public void close() {
		assert (isOpen);

		synchronized (eventLoop.threadLock) {
			stream.drain();
			stream.disconnect();
		}

		try {
			semaphore.acquire();
		} catch (InterruptedException e) {
			// throw new LineUnavailableException("unable to prepare
			// stream");
		}

	}

	public int getBufferSize() {
		// FIXME!
		return 10000;
	}

	public AudioFormat getFormat() {
		return currentFormat;
	}

	public void setDefaultFormat(AudioFormat format) {
		for (AudioFormat aFormat : supportedFormats) {
			if (format.matches(aFormat)) {
				defaultFormat = format;
			}
		}
	}

	public int getFramePosition() {
		return (int) currentFramePosition;
	}

	public float getLevel() {
		return AudioSystem.NOT_SPECIFIED;
	}

	public long getLongFramePosition() {
		return currentFramePosition;
	}

	public long getMicrosecondPosition() {

		float frameRate = currentFormat.getFrameRate();
		float time = currentFramePosition / frameRate; // seconds
		long microseconds = (long) (time * 1000);
		return microseconds;
	}

	public boolean isActive() {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean isRunning() {
		// TODO Auto-generated method stub
		return false;
	}

	public Control getControl(Type control) {
		if (!isOpen) {
			throw new IllegalArgumentException(
					"Controls only supported when line is open");
		}

		for (int i = 0; i < controls.length; i++) {
			if (controls[i].getType().getClass() == control.getClass()) {
				return controls[i];
			}
		}
		throw new IllegalArgumentException("Unsupported control type");
	}

	public Control[] getControls() {
		if (isOpen) {
			return controls;
		} else {
			return new Control[] {};
		}

	}

	public javax.sound.sampled.Line.Info getLineInfo() {
		return new DataLine.Info(SourceDataLine.class, supportedFormats, 0,
				100000);
	}

	public boolean isControlSupported(Type control) {
		for (Control myControl : controls) {
			if (myControl.getType().getClass() == control.getClass()) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void drain() {
		Operation operation;
		Operation.State operationState;
		synchronized (eventLoop.threadLock) {
			operation = stream.drain();
			operationState = operation.getState();
		}

		// FIXME need to find a way to do a wait than a busy loop
		while (operationState != Operation.State.Done) {
			synchronized (eventLoop.threadLock) {
				operationState = operation.getState();
			}
		}
		operation.releaseReference();

	}

	@Override
	public void flush() {
		Operation operation;
		Operation.State operationState;
		synchronized (eventLoop.threadLock) {
			operation = stream.flush();
			operationState = operation.getState();
		}
		// FIXME need to find a way to do a wait than a busy loop
		while (operationState != Operation.State.Done) {
			synchronized (eventLoop.threadLock) {
				operationState = operation.getState();
			}
		}

		operation.releaseReference();

	}

	private void fireLineEvent(LineEvent e) {
		for (LineListener lineListener : lineListeners) {
			lineListener.update(e);
		}
	}

	protected EventLoop getEventLoop() {
		return this.eventLoop;
	}

	public Stream getStream() {
		return stream;
	}

}
