/* PulseAudioTargetDataLine.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
*/

package org.classpath.icedtea.pulseaudio;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Semaphore;

import javax.sound.sampled.*;
import javax.sound.sampled.AudioFormat.Encoding;
import javax.sound.sampled.Control.Type;
import javax.sound.sampled.Port.Info;

public class PulseAudioTargetDataLine implements TargetDataLine {

	
	
	protected boolean isOpen = false;
	protected boolean isPaused = false;

	private AudioFormat[] supportedFormats = null;
	private AudioFormat currentFormat = null;
	private AudioFormat defaultFormat = null;

	private List<LineListener> lineListeners = new LinkedList<LineListener>();
	
	private List<StreamListener> streamListeners = new ArrayList<StreamListener>();
	
	private String streamName = "Java Stream";
	private static final int DEFAULT_BUFFER_SIZE = 1000;
	private static final String PULSEAUDIO_FORMAT_KEY = "PulseAudioFormatKey";
	
	private EventLoop eventLoop = null;
	
	private Semaphore semaphore = new Semaphore(0);
	
	@SuppressWarnings("unused")
	private long streamPointer;

	static {
		try {
			String library = new java.io.File(".").getCanonicalPath()	+ java.io.File.separatorChar
					+ System.mapLibraryName("pulse-java");
			System.out.println(library);
			System.load(library);
		} catch (IOException e) {
			assert ("Loading failed".endsWith("library"));
		}
	}
	
	private native void native_open(long contextPointer, String streamName,
			String encoding, int sampleRate, int channels, int bufferSize);
	
	private native void native_start();
	
	private native int native_get_readable_size();
	
	private native void native_close();
	
	private native int native_read(byte[] array, int remaininglength, int position);
	
	public PulseAudioTargetDataLine(EventLoop eventLoop, AudioFormat[] formats, AudioFormat defaultFormat) {
		supportedFormats = formats;
		this.eventLoop = eventLoop;
		this.lineListeners = new ArrayList<LineListener>();
		this.defaultFormat = defaultFormat; 
		this.currentFormat = defaultFormat;

	}

	public void open(AudioFormat format, int bufferSize)
			throws LineUnavailableException {
		System.out.println("OPEn CALLED");
		if (isOpen) {
			throw new IllegalStateException("Line is already open");
		}

		// ignore suggested buffer size

		for (AudioFormat myFormat : supportedFormats) {
			if (format.matches(myFormat)) {
				native_open(eventLoop.getContextPointer(), streamName,
						(String) myFormat.getProperty(PULSEAUDIO_FORMAT_KEY),
						(int) format.getSampleRate(), format.getChannels(),
						bufferSize);
				currentFormat = format;
				isOpen = true;
			}
		}
		// no matches found
		if (!isOpen) {
			throw new IllegalArgumentException("Invalid format");
		}

		StreamListener openCloseListener = new StreamListener() {

			@Override
			public void update(StreamEvent e) {
				if (e.getType() == StreamEvent.Type.READY) {
					fireLineEvent(new LineEvent(PulseAudioTargetDataLine.this,
							LineEvent.Type.OPEN, AudioSystem.NOT_SPECIFIED));
					System.out.println("IN HERE");
					semaphore.release();
				} else if (e.getType() == StreamEvent.Type.TERMINATED
						|| e.getType() == StreamEvent.Type.FAILED) {
					fireLineEvent((new LineEvent(PulseAudioTargetDataLine.this,
							LineEvent.Type.CLOSE, AudioSystem.NOT_SPECIFIED)));
					semaphore.release();
				}
			}

		};

		addStreamListener(openCloseListener);


		
		synchronized (eventLoop.threadLock) {

			native_start();
		}

		try {
			semaphore.acquire();
		} catch (InterruptedException e) {
			// throw new LineUnavailableException("unable to prepare
			// stream");
		}
	}

	public void open(AudioFormat format) throws LineUnavailableException {
		open(format, DEFAULT_BUFFER_SIZE);

	}

	public void open() throws LineUnavailableException {
		// pick a random format
		if (defaultFormat == null) {
			defaultFormat = new AudioFormat(Encoding.PCM_UNSIGNED, 22050, 8, 2,
					2, AudioSystem.NOT_SPECIFIED, false);
		}

		open(defaultFormat, DEFAULT_BUFFER_SIZE);
	}


	@Override
	public int read(byte[] data, int offset, int length) {
		int frameSize = currentFormat.getFrameSize();
		
		if (length % frameSize != 0) {
			throw new IllegalArgumentException(
					"amount of data to write does not represent an integral number of frames");
		}
		
		if (length < 0) {
			throw new IllegalArgumentException("length is negative");
		}

		if (length + offset > data.length) {
			throw new ArrayIndexOutOfBoundsException(length + offset);
		}

		int position = offset;
		int remainingLength = length;
		int availableSize;

		int sizeRead = 0;

		while (remainingLength != 0) {

			synchronized (eventLoop.threadLock) {
				availableSize = available();
				int toRead = native_read(data, remainingLength,  position);

				sizeRead += toRead;
				position += toRead;
				remainingLength -= toRead;
				

			}
		}

		// all the data should have been played by now
		assert (sizeRead == length);
		//currentFramePosition += (sizeWritten/getFormat().getFrameSize());
		/*
		 * FIXME when the stream is flushed() etc, instead of returning length
		 * this should unblock and return the the size of data written so far
		 */
		return sizeRead;
	
	}

	private native void readFromStream(byte[] b, int off, int len);

	public void start() {
		if (isPaused) {
			resumeStream();
			isPaused = false;
		} /*else {
			startStream();
		}*/

		/*for (LineListener l : listeners) {
			l.update(new LineEvent(this, LineEvent.Type.START, 0));
		}*/
	}

	public void stop() {
		//pauseStream();
		isPaused = true;

	}

	private native void startStream();

	private native void pauseStream();

	private native void resumeStream();

	@Override
	public void drain() {
		// TODO Auto-generated method stub

	}

	@Override
	public void flush() {
		// TODO Auto-generated method stub

	}

	public void addLineListener(LineListener listener) {
		lineListeners.add(listener);
	}

	public void removeLineListener(LineListener listener) {
		lineListeners.remove(listener);
	}

	public boolean isOpen() {
		return isOpen;
	}

	public int available() {
		synchronized (eventLoop.threadLock) {
			return native_get_readable_size();
		}
	}

	public void close() {
		assert (isOpen);


		synchronized (eventLoop.threadLock) {
			native_close();
		}

		try {
			semaphore.acquire();
			System.out.println("stream closed");
		} catch (InterruptedException e) {
			// throw new LineUnavailableException("unable to prepare
			// stream");
		}

	}

	private native void closeStream();

	public int getBufferSize() {
		// TODO Auto-generated method stub
		return 0;
	}

	public AudioFormat getFormat() {
		return currentFormat;
	}

	public int getFramePosition() {
		// TODO Auto-generated method stub
		return 0;
	}

	public float getLevel() {
		// TODO Auto-generated method stub
		return 0;
	}

	public long getLongFramePosition() {
		// TODO Auto-generated method stub
		return 0;
	}

	public long getMicrosecondPosition() {
		// TODO Auto-generated method stub
		return 0;
	}

	public boolean isActive() {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean isRunning() {
		// TODO Auto-generated method stub
		return false;
	}

	public Control getControl(Type control) {
		// TODO Auto-generated method stub
		return null;
	}

	public Control[] getControls() {
		// TODO Auto-generated method stub
		return null;
	}

	public javax.sound.sampled.Line.Info getLineInfo() {
		return new Line.Info(SourceDataLine.class);
	}

	public boolean isControlSupported(Type control) {
		// TODO Auto-generated method stub
		return false;
	}
	
	public void update(int status) {
		synchronized (eventLoop.threadLock) {
			switch (status) {
			case 0:
				fireStreamEvent(new StreamEvent(StreamEvent.Type.UNCONNECTED));
				break;
			case 1:
				fireStreamEvent(new StreamEvent(StreamEvent.Type.CREATING));
				break;
			case 2:
				fireStreamEvent(new StreamEvent(StreamEvent.Type.READY));
				break;
			case 3:
				fireStreamEvent(new StreamEvent(StreamEvent.Type.FAILED));
				break;
			case 4:
				fireStreamEvent(new StreamEvent(StreamEvent.Type.TERMINATED));
				break;
			default:
				assert ("not supposed to happen".indexOf("false") >= 0);
			}
		}
	}

	private void fireLineEvent(LineEvent e) {
		for (LineListener lineListener : lineListeners) {
			lineListener.update(e);
		}
	}

	private void fireStreamEvent(StreamEvent e) {
		synchronized (streamListeners) {
			for (StreamListener streamListener : streamListeners) {
				streamListener.update(e);
			}
		}
	}
	
	
	private void addStreamListener(StreamListener listener) {
		synchronized (streamListeners) {
			this.streamListeners.add(listener);
		}
	}

	private void removeStreamListener(StreamListener listener) {
		synchronized (streamListeners) {
			this.streamListeners.remove(listener);
		}
	}

}