/* jni-common.c
 Copyright (C) 2008 Red Hat, Inc.

 This file is part of IcedTea.

 IcedTea is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as published by
 the Free Software Foundation, version 2.

 IcedTea is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with IcedTea; see the file COPYING.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 02110-1301 USA.

 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
 conditions of the GNU General Public License cover the whole
 combination.

 As a special exception, the copyright holders of this library give you
 permission to link this library with independent modules to produce an
 executable, regardless of the license terms of these independent
 modules, and to copy and distribute the resulting executable under
 terms of your choice, provided that you also meet, for each linked
 independent module, the terms and conditions of the license of that
 module.  An independent module is a module which is not derived from
 or based on this library.  If you modify this library, you may extend
 this exception to your version of the library, but you are not
 obligated to do so.  If you do not wish to do so, delete this
 exception statement from your version.
 */

#include "jni-common.h"

#include <assert.h>

/*
 * Throw an exception by name
 */
void throwByName(JNIEnv* env, const char* name, const char* msg) {
	jclass cls = (*env)->FindClass(env, name);
	if (cls != NULL) {
		(*env)->ThrowNew(env, cls, msg);
		return;
	}
}

jint getJavaIntField(JNIEnv* env, jobject obj, char* fieldName) {
	jclass cls = (*env)->GetObjectClass(env, obj);
	assert(cls);
	jfieldID fid = (*env)->GetFieldID(env, cls, fieldName, "I");
	assert(fid);
	jint value = (*env)->GetIntField(env, obj, fid);
	return value;
}

void setJavaIntField(JNIEnv *env, jobject obj, char *fieldName, jint value) {
	jclass cls = (*env)->GetObjectClass(env, obj);
	assert(cls);
	jfieldID fid =(*env)->GetFieldID(env, cls, fieldName, "I");
	assert(fid);
	(*env)->SetIntField(env, obj, fid, value);
}

jlong getJavaLongField(JNIEnv* env, jobject obj, char* name) {
	jclass cls = (*env)->GetObjectClass(env, obj);
	assert(cls);
	jfieldID fid = (*env)->GetFieldID(env, cls, name, "J");
	assert(fid);
	jint value = (*env)->GetLongField(env, obj, fid);
	return value;

}

void setJavaLongField(JNIEnv* env, jobject obj, char* name, jlong value) {
	jclass cls = (*env)->GetObjectClass(env, obj);
	assert(cls);
	jfieldID fid =(*env)->GetFieldID(env, cls, name, "J");
	assert(fid);
	(*env)->SetLongField(env, obj, fid, value);
}

void callJavaVoidMethod(JNIEnv* env, jobject obj, const char* method_name) {

	jclass cls = (*env)->GetObjectClass(env, obj);
	if (cls == NULL) {
		printf("unable to get class of object");
		return;
	}
	jmethodID mid = (*env)->GetMethodID(env, cls, method_name, "()V");
	if (mid == NULL) {
		printf("unable to get method %s\n", method_name);
		return;

	}
	(*env)->CallVoidMethod(env, obj, mid);

	return;

}

/*
 * 
 * 
 * The functions below are going to cause headaches when porting
 * 
 * 
 */

void* getJavaPointer(JNIEnv* env, jobject obj, char* name) {

	/*
	 * on a 64 bit machine, the value of the long is 64 bits, which should be the pointer
	 * size and everything should be ok
	 * 
	 * So the problem is 32 bit architectures:
	 * 
	 * A reference snippet 
	 * 
	 int a_int = 0x1020;
	 long a_long = 0x102030;
	 long long a_long_long = 0x1020304150607080;
	 printf("size of int: %d\nsize of long: %d\nsize of long long: %d\n", 
	 sizeof(int), sizeof(long), sizeof(long long));
	 printf("int to long long: %X -> %llX\n", a_int, (long long) a_int);
	 printf("long long to int: %llX -> %X\n", a_long_long, (int)a_long_long); 

	 * this shows that long long -> int does a bitwise mask to get the lower order bits only
	 * this is a guess at how it works:
	 *  
	 *   raw data in register is 0x 10203040 50607080
	 *                               32 bits  32 bits
	 * 
	 * we are interested in the 50607080 part of data
	 * 
	 *	Big endian:
	 *    	10	<-- Pointer         <	32bit pointer (for an int)
	 *    	20						<
	 *    	30          			<
	 *    	40                     	<
	 * 		50
	 * 		60
	 * 		70
	 * 		80
	 * (for big endian, will need to do masking and shifting)
	 * 
	 * 
	 * 	Little Endian
	 * 		80	<-- Pointer			< 	32bit pointer (for an int)
	 * 		70						<
	 * 		60						<
	 * 		50						<
	 * 		40
	 * 		30
	 * 		20
	 * 		10
	 * (so little endian works by default!)
	 * 
	 * 
	 */

	/*
	 * jlong = 64 bits
	 * long long = at least 64 bits
	 */

	long long value = getJavaLongField(env, obj, name);

	jPointer result = (jPointer) value;
	return (void*) result;

}

void setJavaPointer(JNIEnv* env, jobject obj, char* name, void* pointer_value) {

	setJavaLongField(env, obj, name, (jPointer) pointer_value);

}

void* convertJavaLongToPointer(jlong pointer) {

	/* this truncates the jlong */
	jPointer p = pointer;
	return (void*) p;

}

jlong convertPointerToJavaLong(void* pointer) {
	return (long) pointer;

}
