/* PulseAudioStreamVolumeControl.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import static org.junit.Assert.assertNotNull;

import java.net.UnknownHostException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;

import junit.framework.JUnit4TestAdapter;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

public class PulseAudioMixerTest {

	PulseAudioMixer selectedMixer;

	public static junit.framework.Test suite() {
		return new JUnit4TestAdapter(PulseAudioMixerTest.class);
	}

	@Before
	public void setUp() throws Exception {
		Mixer.Info mixerInfos[] = AudioSystem.getMixerInfo();
		Mixer.Info selectedMixerInfo = null;
		// int i = 0;
		for (Mixer.Info info : mixerInfos) {
			// System.out.println("Mixer Line " + i++ + ": " + info.getName() +
			// " " + info.getDescription());
			if (info.getName().contains("PulseAudio")) {
				selectedMixerInfo = info;
			}
		}
		assertNotNull(selectedMixerInfo);
		selectedMixer = (PulseAudioMixer) AudioSystem
				.getMixer(selectedMixerInfo);
		assertNotNull(selectedMixer);
		if (selectedMixer.isOpen()) {
			selectedMixer.close();
		}

	}

	@Test
	public void testLocalOpen() throws LineUnavailableException {
		System.out.println("This test tries to open to the local system");
		selectedMixer.openLocal();
	}

	@Test
	public void testLocalOpenAppName() throws LineUnavailableException {
		System.out
				.println("This test tries to connect to the local system while using an application name");
		selectedMixer.openLocal("JunitTest");

	}

	@Test(expected = LineUnavailableException.class)
	public void testRemoteOpenWithInvalidPort() throws UnknownHostException,
			LineUnavailableException {
		System.out
				.println("this test tries to connect to an invalid remote system");
		selectedMixer.openRemote("JUnitTest", "128.0.0.1", 10);

	}

	/*
	 * This test assumes a computer named 'town' is in the network with
	 * pulseaudio listening on port 4173
	 */
	@Test
	public void testRemoteOpenWithValidPort() throws UnknownHostException,
			LineUnavailableException {
		System.out.println("This test tries to connect a valid remote system");
		selectedMixer.openRemote("JUnitTest", "town", 4713);
		selectedMixer.close();
	}

	/*
	 * This test assumes a computer named 'town' is in the network with
	 * pulseaudio listening
	 */
	@Test
	public void testRemoteOpen() throws UnknownHostException,
			LineUnavailableException {
		selectedMixer.openRemote("JUnitTest", "town");
		selectedMixer.close();
	}

	@Test(expected = LineUnavailableException.class)
	public void testInvalidRemoteOpen() throws UnknownHostException,
			LineUnavailableException {
		selectedMixer.openRemote("JUnitTest", "127.0.0.1");
		selectedMixer.close();
	}

	@Test
	public void testSourceLinesExist() throws LineUnavailableException {
		selectedMixer.open();
		Line.Info allLineInfo[] = selectedMixer.getSourceLineInfo();
		int j = 0;
		for (Line.Info lineInfo : allLineInfo) {
			System.out.println("Source Line " + j++ + ": "
					+ lineInfo.getLineClass());
			SourceDataLine sourceDataLine = (SourceDataLine) selectedMixer
					.getLine(lineInfo);
			assertNotNull(sourceDataLine);

		}

	}

	@Test
	public void testOpeningAgain() throws LineUnavailableException,
			UnsupportedOperationException {
		selectedMixer.open();
		selectedMixer.open();
	}

	@Test
	public void testClosingAgain() throws LineUnavailableException,
			UnsupportedOperationException {
		selectedMixer.close();
		selectedMixer.close();
	}

	@Test @Ignore 
	public void testSourceLinesOpenAndClose() throws LineUnavailableException {
		System.out.println("This test checks if source lines open and close");
		selectedMixer.open();

		/*
		 * FIXME This test currently fails. The mixer returns information about
		 * the line which leaves a lot of things as NOT_SPECIFIED when using
		 * that to do a get line, things match, and the line returned still has
		 * a few parameters as NOT_SPECIFIED and doing an open() on that fails
		 * 
		 */
		Line.Info allLineInfo[] = selectedMixer.getSourceLineInfo();
		for (Line.Info lineInfo : allLineInfo) {
			SourceDataLine sourceDataLine = (SourceDataLine) selectedMixer
					.getLine(lineInfo);
			System.out.println("opening line");
			sourceDataLine.open();
			System.out.println("closing line");
			sourceDataLine.close();
		}
	}

	@Test
	public void testOpenEvent() throws LineUnavailableException {
		LineListener listener = new LineListener() {
			private int called = 0;

			@Override
			public void update(LineEvent event) {
				assert (event.getType() == LineEvent.Type.OPEN);
				called++;
				// assert listener is called exactly once
				Assert.assertEquals(1,called);
			}
		};

		selectedMixer.addLineListener(listener);
		selectedMixer.open();
		selectedMixer.removeLineListener(listener);
		selectedMixer.close();

	}

	@Test
	public void testCloseEvent() throws LineUnavailableException {
		LineListener listener = new LineListener() {
			private int count = 0;

			@Override
			public void update(LineEvent event) {
				Assert.assertTrue(event.getType() == LineEvent.Type.CLOSE);
				count++;
				// assert listener is called exactly once
				Assert.assertEquals(1, count);

			}
		};

		selectedMixer.open();
		selectedMixer.addLineListener(listener);
		selectedMixer.close();
		selectedMixer.removeLineListener(listener);

	}

	@Test
	public void testLongWait() throws LineUnavailableException {
		selectedMixer.open();
		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {
			System.out.println("thread interrupted");
		}

	}

	@After
	public void tearDown() throws Exception {
		if (selectedMixer.isOpen())
			selectedMixer.close();
	}

}
